import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/visitor.dart';
import 'package:build/src/builder/build_step.dart';
import 'package:controller/controller.dart';
import 'package:controller_generator/src/util.dart';
import 'package:source_gen/source_gen.dart';

class ValidatableBuilder extends GeneratorForAnnotation<Validatable> {
  @override
  String generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    var validatorBuilder = ValidatorBuilder();
    element.visitChildren(validatorBuilder);
    return '''
    class ${element.name}\$Validator extends EntityValidator {
      static final ${element.name}\$Validator instance = ${element.name}\$Validator();
    
      ${element.name}\$Validator() : super([${validatorBuilder.validators.join(', ')},]);
    }
    ''';
  }
}

class ValidatorBuilder extends SimpleElementVisitor<void> {
  final List<String> validators = [];

  @override
  void visitFieldElement(FieldElement element) {
    if (!element.isStatic && !element.isPrivate) {
      var validatorsForProperty = <String>[];
      validatorsForProperty.addAll(_customValidators(element));
      validatorsForProperty.addAll(_nestedValidator(element));
      if (validatorsForProperty.isNotEmpty) {
        validators.add("EntityPropertyValidator('${element.name}', "
            "[${validatorsForProperty.join(',')}], "
            '(entity) => entity.${element.name},)');
      }
    }
  }

  List<String> _customValidators(FieldElement element) => element.metadata
      .where((metadata) =>
          isType(metadata.computeConstantValue().type, PropertyValidator))
      .map((metadata) => metadata.toSource().substring(1))
      .toList();

  List<String> _nestedValidator(FieldElement element) =>
      hasMeta(element.type.element, Validatable)
          ? ['${element.type.element.name}\$Validator.instance']
          : [];
}
